/// Copyright (c) 2019 Razeware LLC
///
/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this software and associated documentation files (the "Software"), to deal
/// in the Software without restriction, including without limitation the rights
/// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
/// copies of the Software, and to permit persons to whom the Software is
/// furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// Notwithstanding the foregoing, you may not use, copy, modify, merge, publish,
/// distribute, sublicense, create a derivative work, and/or sell copies of the
/// Software in any work that is designed, intended, or marketed for pedagogical or
/// instructional purposes related to programming, coding, application development,
/// or information technology.  Permission for such use, copying, modification,
/// merger, publication, distribution, sublicensing, creation of derivative works,
/// or sale is expressly withheld.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
/// THE SOFTWARE.

import Foundation

// MARK: - Testing Integer Multiples

let firstNumber = 4
let secondNumber = 2
if firstNumber.isMultiple(of: secondNumber) {
  print("\(secondNumber) * \(firstNumber / secondNumber) = \(firstNumber)")
}

// MARK: - Escaping Raw Strings

let raw = #"You can create "raw"\"plain" strings in Swift 5."#
let multiline = #"""
                You can create """raw"""\""""plain"""" strings
                on multiple lines
                in Swift 5.
                """#

let track = "Nothing Else Matters"
print(#"My favorite tune\song is \#(track)."#)

let hashtag = ##"You can use the Swift "hashtag"#swift in Swift 5."##

let versions = "3 3.1 4 4.1 4.2 5"
let range = NSRange(versions.startIndex..., in: versions)
let regex = try! NSRegularExpression(pattern: #"\d\.\d"#)
let minorVersions = regex.matches(in: versions, range: range)
minorVersions.forEach { print(versions[Range($0.range, in: versions)!]) }

// MARK: - New Character Properties

let id = "ID10"
var digits = 0
id.forEach { digits += $0.isNumber ? 1 : 0 }
print("Id has \(digits) digits.")

// MARK: - New Unicode Scalar Properties

let username = "bond007"
var letters = 0
username.unicodeScalars.forEach { letters += $0.properties.isAlphabetic ? 1 : 0 }
print("Username has \(letters) letters.")

// MARK: - Removing Subsequences

extension Sequence {
  func remove(_ s: String) -> [Element] {
    guard let n = Int(s) else {
      return dropLast()
    }
    return dropLast(n)
  }
}

let sequence = [5, 2, 7, 4]
sequence.remove("2")
sequence.remove("two")

// MARK: - Compacting Dictionaries

let students = ["Oana": "10", "Nori": "ten"]
let mapStudents = students.compactMapValues(Int.init)

// MARK: - Renaming Dictionary Literals

let pets: KeyValuePairs = ["dog": "Sclip", "cat": "Peti"]

// MARK: - Numeric Protocol Updates

struct Vector {
  let x, y: Int
  
  init(_ x: Int, _ y: Int) {
    self.x = x
    self.y = y
  }
}

// MARK: - AdditiveArithmetic

extension Vector: AdditiveArithmetic {
  static var zero: Vector {
    return Vector(0, 0)
  }
  
  static func +(lhs: Vector, rhs: Vector) -> Vector {
    return Vector(lhs.x + rhs.x, lhs.y + rhs.y)
  }
  
  static func +=(lhs: inout Vector, rhs: Vector) {
    lhs = lhs + rhs
  }
  
  static func -(lhs: Vector, rhs: Vector) -> Vector {
    return Vector(lhs.x - rhs.x, lhs.y - rhs.y)
  }
  
  static func -=(lhs: inout Vector, rhs: Vector) {
    lhs = lhs - rhs
  }
}

// MARK: - CustomStringConvertible

extension Vector: CustomStringConvertible {
  var description: String {
    return "(\(x) \(y))"
  }
}

var first = Vector(1, 2)
let second = Vector(3, 4)
let third = first + second
first += second
let fourth = first - second
first -= second

// MARK: - String Interpolation Updates

var interpolation = DefaultStringInterpolation(literalCapacity: 7, interpolationCount: 1)
let language = "Swift"
interpolation.appendLiteral(language)
let space = " "
interpolation.appendLiteral(space)
let version = 5
interpolation.appendInterpolation(version)
let string = String(stringInterpolation: interpolation)

// MARK: - Handling Future Enumeration Cases

enum Post {
  case tutorial, article, podcast, screencast, course
}

func readPost(_ post: Post) -> String {
  switch post {
    case .tutorial:
      return "You are reading a tutorial."
    case .article:
      return "You are reading an article."
    @unknown default:
      return "You are reading a blog post."
  }
}

let screencast = Post.screencast
readPost(screencast)
let course = Post.course
readPost(course)
let podcast = Post.podcast
readPost(podcast)

// MARK: - Adding Result to the Standard Library

enum ConnectionError: Error {
  case noNetwork, noDatabase
}

let networkSuccess = Result<String, ConnectionError>.success("Network connected!")
let databaseSuccess = Result<String, ConnectionError>.success("Database connected!")
let networkFailure = Result<String, ConnectionError>.failure(.noNetwork)
let databaseFailure = Result<String, ConnectionError>.failure(.noDatabase)
let sameSuccess = networkSuccess == databaseSuccess
let sameFailure = networkFailure == databaseFailure
let success: Set = [networkSuccess, databaseSuccess]
let failure: Set = [networkFailure, databaseFailure]
let successDictionary = [networkSuccess: try! networkSuccess.get(), databaseSuccess: try! databaseSuccess.get()]
let failureDictionary = [networkFailure: ConnectionError.noNetwork, databaseFailure: ConnectionError.noDatabase]

// MARK: - Conforming Never to Equatable and Hashable Protocols

let alwaysSucceeds = Result<String, Never>.success("Network connected!")
let neverFails = Result<String, Never>.success("Database connected!")
let alwaysFails = Result<Never, ConnectionError>.failure(.noNetwork)
let neverSucceeds = Result<Never, ConnectionError>.failure(.noDatabase)
let sameValue = alwaysSucceeds == neverFails
let sameError = alwaysFails == neverSucceeds
let alwaysSuccess: Set = [alwaysSucceeds, neverFails]
let alwaysFailure: Set = [alwaysFails, neverSucceeds]
let alwaysSuccessDictionary = [alwaysSucceeds: try! alwaysSucceeds.get(), neverFails: try! neverFails.get()]
let alwaysFailureDictionary = [alwaysFails: ConnectionError.noNetwork, neverSucceeds: ConnectionError.noDatabase]

// MARK: - Dynamically Callable Types

@dynamicCallable
class DynamicFeatures {
  func dynamicallyCall(withArguments params: [Int]) -> Int? {
    guard !params.isEmpty else {
      return nil
    }
    return params.reduce(0, +)
  }
  
  func dynamicallyCall(withKeywordArguments params: KeyValuePairs<String, Int>) -> Int? {
    guard !params.isEmpty else {
      return nil
    }
    return params.reduce(0) { $1.key.isEmpty ? $0 : $0 + $1.value }
  }
}

let features = DynamicFeatures()
features()
features(3, 4, 5)
features(first: 3, 4, second: 5)

// MARK: - Making Codable Ranges

let temperature = 0...10
let encoder = JSONEncoder()
let data = try! encoder.encode(temperature)
let decoder = JSONDecoder()
let temperatureRange = try! decoder.decode(ClosedRange<Int>.self, from: data)

// MARK: - Flattening Nested Optionals

extension Int {
  enum DivisionError: Error {
    case divisionByZero
  }
  
  func divideBy(_ number: Int) throws -> Int {
    guard number != 0 else {
      throw DivisionError.divisionByZero
    }
    return self / number
  }
}

let number: Int? = 10
let division = try? number?.divideBy(2)
if let division = division {
  print(division)
}

// MARK: - Identity Key Paths

class Tutorial {
  let title: String
  let author: String
  
  init(title: String, author: String) {
    self.title = title
    self.author = author
  }
}

var tutorial = Tutorial(title: "What's New in Swift 5.0?", author: "Cosmin Pupaza")
tutorial[keyPath: \.self] = Tutorial(title: "What's New in Swift 5?", author: "Cosmin Pupăză")

// MARK: - Initializing Literals Through Coercion

let value = UInt64(0xFFFF_FFFF_FFFF_FFFF)

// MARK: - Build Configuration Updates

let favoriteNumber = 10
var evenNumber = true

#if swift(<5)
  evenNumber = favoriteNumber % 2 == 0
#else
  evenNumber = favoriteNumber.isMultiple(of: 2)
#endif

#if compiler(<5)
  evenNumber = favoriteNumber % 2 == 0
#else
  evenNumber = favoriteNumber.isMultiple(of: 2)
#endif

// MARK: - Using Variadic Parameters for Enumeration Cases with Associated Values

enum BlogPost {
  case tutorial([String])
  case article([String])
}

// MARK: - Deprecating String Index Encoded Offsets

let swiftVersion = "Swift 5"
let offset = swiftVersion.endIndex.utf16Offset(in: swiftVersion)









































































